from __future__ import absolute_import

import logging
import typing

import pyspark.sql as spark

from dbnd_spark.spark_targets.spark_histograms import SparkHistograms
from targets.value_meta import ValueMeta
from targets.values.builtins_values import DataValueType


if typing.TYPE_CHECKING:
    from targets.value_meta import ValueMetaConf

logger = logging.getLogger(__name__)


class SparkDataFrameValueType(DataValueType):
    type = spark.DataFrame
    type_str = "Spark.DataFrame"
    support_merge = False

    config_name = "spark_dataframe"

    def to_signature(self, x):
        id = "rdd-%s-at-%s" % (x.rdd.id(), x.rdd.context.applicationId)
        return id

    def to_preview(self, df, preview_size):  # type: (spark.DataFrame, int) -> str
        return (
            df.limit(1000)
            .toPandas()
            .to_string(index=False, max_rows=20, max_cols=1000)[:preview_size]
        )

    def get_value_meta(self, value, meta_conf):
        # type: (spark.DataFrame, ValueMetaConf) -> ValueMeta

        if meta_conf.log_schema:
            data_schema = {
                "type": self.type_str,
                "columns": list(value.schema.names),
                "dtypes": {f.name: str(f.dataType) for f in value.schema.fields},
            }
        else:
            data_schema = None

        if meta_conf.log_preview:
            data_preview = self.to_preview(value, meta_conf.get_preview_size())
        else:
            data_preview = None

        if meta_conf.log_size:
            data_schema = data_schema or {}
            rows = value.count()
            data_dimensions = (rows, len(value.columns))
            data_schema.update(
                {
                    "size": int(rows * len(value.columns)),
                    "shape": (rows, len(value.columns)),
                }
            )
        else:
            data_dimensions = None

        if meta_conf.log_histograms or meta_conf.log_stats:
            spark_histograms = SparkHistograms(value, meta_conf)
            df_stats, histogram_dict = spark_histograms.get_histograms_and_stats()
            hist_sys_metrics = spark_histograms.system_metrics
        else:
            df_stats, histogram_dict = {}, {}
            hist_sys_metrics = None

        return ValueMeta(
            value_preview=data_preview,
            data_dimensions=data_dimensions,
            data_schema=data_schema,
            data_hash=self.to_signature(value),
            descriptive_stats=df_stats,
            histogram_system_metrics=hist_sys_metrics,
            histograms=histogram_dict,
        )

    def support_fast_count(self, target):
        from targets import FileTarget

        if not isinstance(target, FileTarget):
            return False
        from targets.target_config import FileFormat

        return target.config.format == FileFormat.parquet



import logging
import sys

from dbnd._core.utils.seven import import_errors


logger = logging.getLogger(__name__)


def get_spark_session():
    from pyspark.sql import SparkSession

    return SparkSession.builder.getOrCreate()


def has_spark_session():
    if not has_pyspark_imported():
        logger.debug("Spark not found in modules")
        return False

    try:
        from pyspark.sql import SparkSession

        logger.debug(
            "Spark session SparkSession._instantiatedSession %s",
            SparkSession._instantiatedSession,
        )
        return SparkSession._instantiatedSession is not None
    except import_errors:
        return False


def has_pyspark_imported():
    if "pyspark" not in sys.modules:
        logger.debug("Spark not found in modules")
        return False
    return True



# Partially copied from sparkmagic package
# Copyright (c) 2015  aggftw@gmail.com
# Distributed under the terms of the Modified BSD License.
import json
import logging
import time

from dbnd._core.errors import DatabandError
from dbnd._core.utils.http.reliable_http_client import ReliableHttpClient
from dbnd._core.utils.http.retry_policy import get_retry_policy


logger = logging.getLogger(__name__)

BATCH_RUNNING_STATES = ["starting", "not_started", "running", "recovering"]
BATCH_ERROR_STATES = ["error", "dead"]


class LivyBatchClient(object):
    """A Livy-specific Http client which wraps the normal ReliableHttpClient. Propagates
    HttpClientExceptions up."""

    def __init__(self, http_client, endpoint):
        self.endpoint = endpoint
        self._http_client = http_client

    @staticmethod
    def from_endpoint(endpoint):
        headers = {"Content-Type": "application/json"}
        custom_headers = {}

        headers.update(custom_headers)
        retry_policy = LivyBatchClient._get_retry_policy()
        return LivyBatchClient(
            ReliableHttpClient(endpoint, headers, retry_policy), endpoint
        )

    def post_batch(self, properties):
        return self._http_client.post("/batches", [201], properties).json()

    def get_batch(self, batch_id):
        return self._http_client.get(self._batch_url(batch_id), [200]).json()

    def get_batch_state(self, batch_id):
        return self._http_client.get(self._batch_url(batch_id), [200]).json()

    def get_all_batch_logs(self, batch_id, from_line=0):
        return self._http_client.get(
            self._batch_url(batch_id) + "/log?from=%s" % from_line, [200]
        ).json()

    def get_batches(self):
        return self._http_client.get("/batches", [200]).json()

    def delete_batch(self, batch_id):
        self._http_client.delete(self._batch_url(batch_id), [200, 404])

    def get_headers(self):
        return self._http_client.get_headers()

    @staticmethod
    def _batch_url(batch_id):
        return "/batches/{}".format(batch_id)

    @staticmethod
    def _get_retry_policy():
        return get_retry_policy("LivyBatch'")

        # Function to help track the progress of the scala code submitted to Apache Livy

    def _print_status(self, response):
        logger.info

    def _default_status_reporter(self, batch_response):
        logger.info(
            "Batch status: %s", json.dumps(batch_response, indent=4, sort_keys=True)
        )

    def track_batch_progress(self, batch_id, status_reporter=None):
        status_reporter = status_reporter or self._default_status_reporter

        # Poll the status of the submitted scala code

        current_line = 0
        while True:
            batch_response = self.get_batch(batch_id)
            batch_status = batch_response["state"]
            status_reporter(batch_response)

            # logging the logs
            lines = self.get_all_batch_logs(batch_id, from_line=current_line)["log"]
            for line in lines:
                logger.info(line)
            current_line += len(lines)
            if batch_status.lower() not in BATCH_RUNNING_STATES:
                break
            time.sleep(10)

        status_reporter(batch_response)
        if batch_status.lower() in BATCH_ERROR_STATES:
            logger.info("Batch exception: see logs")
            raise DatabandError("Batch Status: " + batch_status)
        logger.info("Batch Status: " + batch_status)
